"""
Configuration module for schema section test.
"""

from dataclasses import dataclass
from typing import Any, Callable, List, Optional


@dataclass
class OpenAPITestConfig:
    """Configuration dataclass for schema section test."""

    case_tester: Optional[Callable[[str], None]] = None
    ignore_case: Optional[List[str]] = None
    validators: Any = None
    reference: str = "root"
    http_message: str = "response"



"""
Utils Module - this file contains utility functions used in multiple places.
"""

from __future__ import annotations

import json
from copy import deepcopy
from itertools import chain, combinations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Iterator, Sequence


def merge_objects(dictionaries: Sequence[dict[str, Any]]) -> dict[str, Any]:
    """
    Deeply merge objects.
    """
    output: dict[str, Any] = {}
    for dictionary in dictionaries:
        for key, value in dictionary.items():
            if key not in output:
                output[key] = value
                continue
            current_value = output[key]
            if isinstance(current_value, list) and isinstance(value, list):
                output[key] = list(chain(output[key], value))
                continue
            if isinstance(current_value, dict) and isinstance(value, dict):
                output[key] = merge_objects([current_value, value])
                continue
    return output


def normalize_schema_section(schema_section: dict[str, Any]) -> dict[str, Any]:
    """
    Remove allOf and handle edge uses of oneOf.
    """
    output: dict[str, Any] = deepcopy(schema_section)
    if output.get("allOf"):
        all_of = output.pop("allOf")
        output = {**output, **merge_objects(all_of)}
    if output.get("oneOf") and all(item.get("enum") for item in output["oneOf"]):
        # handle the way drf-spectacular is doing enums
        one_of = output.pop("oneOf")
        output = {**output, **merge_objects(one_of)}
    for key, value in output.items():
        if isinstance(value, dict):
            output[key] = normalize_schema_section(value)
        elif isinstance(value, list):
            output[key] = [
                normalize_schema_section(entry) if isinstance(entry, dict) else entry
                for entry in value
            ]
    return output


def serialize_schema_section_data(data: dict[str, Any]) -> str:
    return json.dumps(data, indent=4, default=str)


def lazy_combinations(options_list: Sequence[dict[str, Any]]) -> Iterator[dict]:
    """
    Lazily evaluate possible combinations.
    """
    for i in range(2, len(options_list) + 1):
        for combination in combinations(options_list, i):
            yield merge_objects(combination)


def serialize_json(func):
    def wrapper(*args, **kwargs):
        # import pdb; pdb.set_trace()
        data = kwargs.get("data")
        content_type = kwargs.get("content_type")
        if data and content_type == "application/json":
            try:
                kwargs["data"] = json.dumps(data)
            except (TypeError, OverflowError):
                kwargs["data"] = data
        return func(*args, **kwargs)

    return wrapper



"""Exceptions Module"""


class DocumentationError(AssertionError):
    """
    Custom exception raised when package tests fail.
    """

    pass


class CaseError(DocumentationError):
    """
    Custom exception raised when items are not cased correctly.
    """

    def __init__(self, key: str, case: str, expected: str) -> None:
        super().__init__(
            f"The response key `{key}` is not properly {case}. Expected value: {expected}"
        )


class OpenAPISchemaError(Exception):
    """
    Custom exception raised for invalid schema specifications.
    """

    pass


class UndocumentedSchemaSectionError(OpenAPISchemaError):
    """
    Subset of OpenAPISchemaError, raised when we cannot find a single schema section.
    """

    pass



# pylint: disable=R0903
"""
Module that contains the factory to create response handlers.
"""

from typing import TYPE_CHECKING, Union

from rest_framework.response import Response

from openapi_tester.response_handler import (
    DjangoNinjaResponseHandler,
    DRFResponseHandler,
)

if TYPE_CHECKING:
    from django.http.response import HttpResponse

    from openapi_tester.response_handler import ResponseHandler


class ResponseHandlerFactory:
    """
    Response Handler Factory: this class is used to create a response handler
    instance for both DRF and Django HTTP (Django Ninja) responses.
    """

    @staticmethod
    def create(
        *request_args, response: Union[Response, "HttpResponse"], **kwargs
    ) -> "ResponseHandler":
        if isinstance(response, Response):
            return DRFResponseHandler(response=response)
        return DjangoNinjaResponseHandler(*request_args, response=response, **kwargs)



"""Django OpenAPI Schema Tester"""

from .case_testers import is_camel_case, is_kebab_case, is_pascal_case, is_snake_case
from .clients import OpenAPIClient
from .loaders import (
    BaseSchemaLoader,
    DrfSpectacularSchemaLoader,
    DrfYasgSchemaLoader,
    StaticSchemaLoader,
)
from .schema_tester import SchemaTester

__all__ = [
    "BaseSchemaLoader",
    "DrfSpectacularSchemaLoader",
    "DrfYasgSchemaLoader",
    "SchemaTester",
    "StaticSchemaLoader",
    "is_camel_case",
    "is_kebab_case",
    "is_pascal_case",
    "is_snake_case",
    "OpenAPIClient",
]



"""
This module contains the concrete response handlers for both DRF and Django Ninja responses.
"""

import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from django.http.response import HttpResponse
    from rest_framework.response import Response


@dataclass
class GenericRequest:
    """Generic request class for both DRF and Django Ninja."""

    path: str
    method: str
    data: dict = field(default_factory=dict)
    headers: dict = field(default_factory=dict)


class ResponseHandler(ABC):
    """
    This class is used to handle the response and request data
    from both DRF and Django HTTP (Django Ninja) responses.
    """

    def __init__(self, response: Union["Response", "HttpResponse"]) -> None:
        self._response = response

    @property
    def response(self) -> Union["Response", "HttpResponse"]:
        return self._response

    @property
    @abstractmethod
    def request(self) -> GenericRequest: ...

    @property
    @abstractmethod
    def data(self) -> Optional[dict]: ...


class DRFResponseHandler(ResponseHandler):
    """
    Handles the response and request data from DRF responses.
    """

    def __init__(self, response: "Response") -> None:
        super().__init__(response)

    @property
    def data(self) -> Optional[dict]:
        return self.response.json() if self.response.data is not None else None  # type: ignore[attr-defined]

    @property
    def request(self) -> GenericRequest:
        return GenericRequest(
            path=self.response.renderer_context["request"].path,  # type: ignore[attr-defined]
            method=self.response.renderer_context["request"].method,  # type: ignore[attr-defined]
            data=self.response.renderer_context["request"].data,  # type: ignore[attr-defined]
            headers=self.response.renderer_context["request"].headers,  # type: ignore[attr-defined]
        )


class DjangoNinjaResponseHandler(ResponseHandler):
    """
    Handles the response and request data from Django Ninja responses.
    """

    def __init__(
        self, *request_args, response: "HttpResponse", path_prefix: str = "", **kwargs
    ) -> None:
        super().__init__(response)
        self._request_method = request_args[0]
        self._request_path = f"{path_prefix}{request_args[1]}"
        self._request_data = self._build_request_data(request_args[2])
        self._request_headers = kwargs

    @property
    def data(self) -> Optional[dict]:
        return self.response.json() if self.response.content else None  # type: ignore[attr-defined]

    @property
    def request(self) -> GenericRequest:
        return GenericRequest(
            path=self._request_path,
            method=self._request_method,
            data=self._request_data,
            headers=self._request_headers,
        )

    def _build_request_data(self, request_data: Any) -> dict:
        try:
            return json.loads(request_data)
        except (json.JSONDecodeError, TypeError, ValueError):
            return {}



"""Constants module"""

OPENAPI_PYTHON_MAPPING = {
    "boolean": bool.__name__,
    "string": str.__name__,
    "file": str.__name__,
    "array": list.__name__,
    "object": dict.__name__,
    "integer": int.__name__,
    "number": f"{int.__name__} or {float.__name__}",
}

NUMERIC_FORMATS = ["number", "float", "double"]

INTERNET_PROTOCOLS = ["ipv4", "ipv6", "email"]

# Validation errors
VALIDATE_FORMAT_ERROR = (
    'Expected: {article} "{format}" formatted "{type}" value\n\nReceived: {received}'
)
VALIDATE_PATTERN_ERROR = (
    'The string "{data}" does not match the specified pattern: {pattern}'
)
INVALID_PATTERN_ERROR = "String pattern is not valid regex: {pattern}"
VALIDATE_ENUM_ERROR = "Expected: a member of the enum {enum}\n\nReceived: {received}"
VALIDATE_TYPE_ERROR = 'Expected: {article} "{type}" type value\n\nReceived: {received}'
VALIDATE_MULTIPLE_OF_ERROR = "The value {data} should be a multiple of {multiple}"
VALIDATE_MINIMUM_ERROR = (
    "The value {data} is lower than the specified minimum of {minimum}"
)
VALIDATE_MAXIMUM_ERROR = (
    "The value {data} exceeds the maximum allowed value of {maximum}"
)
VALIDATE_MIN_LENGTH_ERROR = 'The length of "{data}" is shorter than the specified minimum length of {min_length}'
VALIDATE_MAX_LENGTH_ERROR = (
    'The length of "{data}" exceeds the specified maximum length of {max_length}'
)
VALIDATE_MIN_ARRAY_LENGTH_ERROR = "The length of the array {data} is shorter than the specified minimum length of {min_length}"
VALIDATE_MAX_ARRAY_LENGTH_ERROR = "The length of the array {data} exceeds the specified maximum length of {max_length}"
VALIDATE_MINIMUM_NUMBER_OF_PROPERTIES_ERROR = "The number of properties in {data} is fewer than the specified minimum number of properties of {min_length}"
VALIDATE_MAXIMUM_NUMBER_OF_PROPERTIES_ERROR = "The number of properties in {data} exceeds the specified maximum number of properties of {max_length}"
VALIDATE_UNIQUE_ITEMS_ERROR = "The array {data} must contain unique items only"
VALIDATE_NONE_ERROR = "A property received a null value in the {http_message} data, but is a non-nullable object in the schema definition"
VALIDATE_MISSING_KEY_ERROR = (
    "The following property was found in the schema definition, "
    'but is missing from the {http_message} data: "{missing_key}"'
)
VALIDATE_EXCESS_KEY_ERROR = (
    "The following property was found in the {http_message} data, "
    'but is missing from the schema definition: "{excess_key}"'
)
VALIDATE_WRITE_ONLY_RESPONSE_KEY_ERROR = 'The following property was found in the response, but is documented as being "writeOnly": "{write_only_key}"'
VALIDATE_ONE_OF_ERROR = "Expected data to match one and only one of the oneOf schema types; found {matches} matches"
VALIDATE_ANY_OF_ERROR = "Expected data to match one or more of the documented anyOf schema types, but found no matches"
UNDOCUMENTED_SCHEMA_SECTION_ERROR = (
    "Error: Unsuccessfully tried to index the OpenAPI schema by `{key}`. {error_addon}"
)
INIT_ERROR = "Unable to configure loader"



"""Schema Tester"""

from __future__ import annotations

import json
import re
from copy import copy
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.validators import URLValidator

from openapi_tester.config import OpenAPITestConfig
from openapi_tester.constants import (
    INIT_ERROR,
    UNDOCUMENTED_SCHEMA_SECTION_ERROR,
    VALIDATE_ANY_OF_ERROR,
    VALIDATE_EXCESS_KEY_ERROR,
    VALIDATE_MISSING_KEY_ERROR,
    VALIDATE_NONE_ERROR,
    VALIDATE_ONE_OF_ERROR,
    VALIDATE_WRITE_ONLY_RESPONSE_KEY_ERROR,
)
from openapi_tester.exceptions import (
    DocumentationError,
    OpenAPISchemaError,
    UndocumentedSchemaSectionError,
)
from openapi_tester.loaders import (
    DrfSpectacularSchemaLoader,
    DrfYasgSchemaLoader,
    StaticSchemaLoader,
    UrlStaticSchemaLoader,
)
from openapi_tester.utils import (
    lazy_combinations,
    normalize_schema_section,
    serialize_schema_section_data,
)
from openapi_tester.validators import (
    validate_enum,
    validate_format,
    validate_max_items,
    validate_max_length,
    validate_max_properties,
    validate_maximum,
    validate_min_items,
    validate_min_length,
    validate_min_properties,
    validate_minimum,
    validate_multiple_of,
    validate_pattern,
    validate_type,
    validate_unique_items,
)

if TYPE_CHECKING:
    from typing import Optional

    from openapi_tester.response_handler import GenericRequest, ResponseHandler


class SchemaTester:
    """Schema Tester: this is the base class of the library."""

    loader: (
        StaticSchemaLoader
        | DrfSpectacularSchemaLoader
        | DrfYasgSchemaLoader
        | UrlStaticSchemaLoader
    )
    validators: list[Callable[[dict, Any], str | None]]

    def __init__(
        self,
        case_tester: Callable[[str], None] | None = None,
        ignore_case: list[str] | None = None,
        schema_file_path: str | None = None,
        validators: list[Callable[[dict, Any], str | None]] | None = None,
        field_key_map: dict[str, str] | None = None,
        path_prefix: str | None = None,
    ) -> None:
        """
        Iterates through an OpenAPI schema object and API response to check that they match at every level.

        :param case_tester: An optional callable that validates schema and response keys
        :param ignore_case: An optional list of keys for the case_tester to ignore
        :schema_file_path: The file path to an OpenAPI yaml or json file. Only passed when using a static schema loader
        :param path_prefix: An optional string to prefix the path of the schema file
        :raises: openapi_tester.exceptions.DocumentationError or ImproperlyConfigured
        """
        self.case_tester = case_tester
        self._path_prefix = path_prefix
        self.ignore_case = ignore_case or []
        self.validators = validators or []

        if schema_file_path is not None:
            try:
                URLValidator()(schema_file_path)
                self.loader = UrlStaticSchemaLoader(
                    schema_file_path, field_key_map=field_key_map
                )
            except ValidationError:
                self.loader = StaticSchemaLoader(
                    schema_file_path, field_key_map=field_key_map
                )
        elif "drf_spectacular" in settings.INSTALLED_APPS:
            self.loader = DrfSpectacularSchemaLoader(field_key_map=field_key_map)
        elif "drf_yasg" in settings.INSTALLED_APPS:
            self.loader = DrfYasgSchemaLoader(field_key_map=field_key_map)
        else:
            raise ImproperlyConfigured(INIT_ERROR)

    @staticmethod
    def get_key_value(
        schema: dict[str, dict], key: str, error_addon: str = "", use_regex=False
    ) -> dict:
        """
        Returns the value of a given key
        """
        try:
            if use_regex:
                compiled_pattern = re.compile(key)
                for key_ in schema.keys():
                    if compiled_pattern.match(key_):
                        return schema[key_]
            return schema[key]
        except KeyError as e:
            raise UndocumentedSchemaSectionError(
                UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(
                    key=key, error_addon=error_addon
                )
            ) from e

    @staticmethod
    def get_status_code(
        schema: dict[str | int, dict], status_code: str | int, error_addon: str = ""
    ) -> dict:
        """
        Returns the status code section of a schema, handles both str and int status codes
        """
        if str(status_code) in schema:
            return schema[str(status_code)]
        if int(status_code) in schema:
            return schema[int(status_code)]
        raise UndocumentedSchemaSectionError(
            UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(
                key=status_code, error_addon=error_addon
            )
        )

    @staticmethod
    def get_schema_type(schema: dict[str, str]) -> str | None:
        if "type" in schema:
            return schema["type"]
        if "properties" in schema or "additionalProperties" in schema:
            return "object"
        return None

    def get_paths_object(self) -> dict[str, Any]:
        schema = self.loader.get_schema()
        paths_object = self.get_key_value(schema, "paths")
        if self._path_prefix:
            paths_object = {
                f"{self._path_prefix}{key}": value
                for key, value in paths_object.items()
            }

        return paths_object

    def get_response_schema_section(
        self, response_handler: ResponseHandler, test_config: OpenAPITestConfig
    ) -> dict[str, Any]:
        """
        Fetches the response section of a schema, wrt. the route, method, status code, and schema version.

        :param response: DRF Response Instance
        :return dict
        """
        response = response_handler.response
        schema = self.loader.get_schema()

        response_method = response_handler.request.method.lower()
        parameterized_path, _ = self.loader.resolve_path(
            response_handler.request.path,
            method=response_method,
        )
        paths_object = self.get_paths_object()

        route_object = self.get_key_value(
            paths_object,
            parameterized_path,
            (
                f"\n\n{test_config.reference}\n\nUndocumented route {parameterized_path}.\n\nDocumented routes: "
                + "\n\t• ".join(paths_object.keys())
            ),
        )

        method_object = self.get_key_value(
            route_object,
            response_method,
            (
                f"\n\n{test_config.reference}"
                f"\n\nUndocumented method: {response_method}."
                "\n\nDocumented methods: "
                f"{[method.lower() for method in route_object.keys() if method.lower() != 'parameters']}."
            ),
        )

        responses_object = self.get_key_value(method_object, "responses")
        status_code_object = self.get_status_code(
            responses_object,
            response.status_code,
            (
                f"\n\n{test_config.reference}"
                f"\n\nUndocumented status code: {response.status_code}."
                f"\n\nDocumented status codes: {list(responses_object.keys())}. "
            ),
        )

        if "openapi" not in schema:
            # openapi 2.0, i.e. "swagger" has a different structure than openapi 3.0 status sub-schemas
            return self.get_key_value(status_code_object, "schema")

        if status_code_object.get("content"):
            content_object = self.get_key_value(
                status_code_object,
                "content",
                (
                    f"\n\n{test_config.reference}"
                    f"\n\nNo content documented for method: {response_method}, path: {parameterized_path}"
                ),
            )
            json_object = self.get_key_value(
                content_object,
                r"^application\/.*json$",
                (
                    f"\n\n{test_config.reference}"
                    "\n\nNo `application/json` responses documented for method: "
                    f"{response_method}, path: {parameterized_path}"
                ),
                use_regex=True,
            )
            return self.get_key_value(json_object, "schema")

        if response_handler.data:
            raise UndocumentedSchemaSectionError(
                UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(
                    key="content",
                    error_addon=(
                        f"\n\nNo `content` defined for this response: {response_method}, path: {parameterized_path}"
                    ),
                )
            )
        return {}

    def get_request_body_schema_section(
        self, request: GenericRequest, test_config: OpenAPITestConfig
    ) -> dict[str, Any]:
        """
        Fetches the request section of a schema.

        :param response: DRF Request Instance
        :return dict
        """
        request_method = request.method.lower()  # request["REQUEST_METHOD"].lower()

        parameterized_path, _ = self.loader.resolve_path(
            request.path, method=request_method
        )

        paths_object = self.get_paths_object()

        route_object = self.get_key_value(
            paths_object,
            parameterized_path,
            (
                f"\n\n{test_config.reference}\n\nUndocumented route {parameterized_path}.\n\nDocumented routes: "
                + "\n\t• ".join(paths_object.keys())
            ),
        )

        method_object = self.get_key_value(
            route_object,
            request_method,
            (
                f"\n\n{test_config.reference}"
                f"\n\nUndocumented method: {request_method}."
                "\n\nDocumented methods: "
                f"{[method.lower() for method in route_object.keys() if method.lower() != 'parameters']}."
            ),
        )

        if request.data:
            if not any(
                "application/json" == request.headers.get(content_type)
                for content_type in ["content_type", "CONTENT_TYPE", "Content-Type"]
            ):
                return {}

            request_body_object = self.get_key_value(
                method_object,
                "requestBody",
                (
                    f"\n\n{test_config.reference}"
                    f"\n\nNo request body documented for method: {request_method}, path: {parameterized_path}"
                ),
            )

            content_object = self.get_key_value(
                request_body_object,
                "content",
                (
                    f"\n\n{test_config.reference}"
                    f"\n\nNo content documented for method: {request_method}, path: {parameterized_path}"
                ),
            )

            json_object = self.get_key_value(
                content_object,
                r"^application\/.*json$",
                (
                    f"\n\n{test_config.reference}"
                    "\n\nNo `application/json` requests documented for method: "
                    f"{request_method}, path: {parameterized_path}"
                ),
                use_regex=True,
            )

            return self.get_key_value(json_object, "schema")

        return {}

    def handle_one_of(
        self,
        schema_section: dict,
        data: Any,
        reference: str,
        test_config: OpenAPITestConfig,
    ) -> None:
        matches = 0
        passed_schema_section_formats = set()
        for option in schema_section["oneOf"]:
            try:
                test_config.reference = f"{test_config.reference}.oneOf"
                self.test_schema_section(
                    schema_section=option,
                    data=data,
                    test_config=test_config,
                )
                matches += 1
                passed_schema_section_formats.add(option.get("format"))
            except DocumentationError:
                continue
        if matches == 2 and passed_schema_section_formats == {"date", "date-time"}:
            # With Django v4, the datetime validator now parses normal
            # date formats successfully, so a oneOf: date // datetime section
            # will succeed twice where it used to succeed once.
            return
        if matches != 1:
            raise DocumentationError(
                f"{VALIDATE_ONE_OF_ERROR.format(matches=matches)}\n\nReference: {reference}.oneOf"
            )

    def handle_any_of(
        self,
        schema_section: dict,
        data: Any,
        reference: str,
        test_config: OpenAPITestConfig,
    ) -> None:
        any_of: list[dict[str, Any]] = schema_section.get("anyOf", [])
        for schema in chain(any_of, lazy_combinations(any_of)):
            test_config.reference = f"{test_config.reference}.anyOf"
            try:
                self.test_schema_section(
                    schema_section=schema,
                    data=data,
                    test_config=test_config,
                )
                return
            except DocumentationError:
                continue
        raise DocumentationError(
            f"{VALIDATE_ANY_OF_ERROR}\n\nReference: {reference}.anyOf"
        )

    def is_openapi_schema(self) -> bool:
        return self.loader.get_schema().get("openapi") is not None

    @staticmethod
    def test_is_nullable(schema_item: dict) -> bool:
        """
        Checks if the item is nullable.

        OpenAPI 3 ref: https://swagger.io/docs/specification/data-models/data-types/#null
        OpenApi 2 ref: https://help.apiary.io/api_101/swagger-extensions/

        :param schema_item: schema item
        :return: whether or not the item can be None
        """
        openapi_schema_3_nullable = "nullable"
        swagger_2_nullable = "x-nullable"
        openapi_schema_3_1_type_nullable = "null"
        if "oneOf" in schema_item:
            one_of: list[dict[str, Any]] = schema_item.get("oneOf", [])
            return any(
                nullable_key in schema and schema[nullable_key]
                for schema in one_of
                for nullable_key in [openapi_schema_3_nullable, swagger_2_nullable]
            )
        if "anyOf" in schema_item:
            any_of: list[dict[str, Any]] = schema_item.get("anyOf", [])
            return any(
                nullable_key in schema and schema[nullable_key]
                for schema in any_of
                for nullable_key in [openapi_schema_3_nullable, swagger_2_nullable]
            )
        if "type" in schema_item and isinstance(schema_item["type"], list):
            types: list[str] = schema_item["type"]
            return openapi_schema_3_1_type_nullable in types

        return any(
            nullable_key in schema_item and schema_item[nullable_key]
            for nullable_key in [openapi_schema_3_nullable, swagger_2_nullable]
        )

    def test_key_casing(
        self,
        key: str,
        case_tester: Callable[[str], None] | None = None,
        ignore_case: list[str] | None = None,
    ) -> None:
        tester = case_tester or getattr(self, "case_tester", None)
        ignore_case = [*self.ignore_case, *(ignore_case or [])]
        if tester and key not in ignore_case:
            tester(key)

    def test_schema_section(
        self,
        schema_section: dict,
        data: Any,
        test_config: OpenAPITestConfig | None = None,
    ) -> None:
        """
        This method orchestrates the testing of a schema section
        """
        test_config = test_config or OpenAPITestConfig()
        if data is None:
            if self.test_is_nullable(schema_section) or not schema_section:
                # If data is None and nullable, we return early
                return
            raise DocumentationError(
                f"{VALIDATE_NONE_ERROR.format(http_message=test_config.http_message)}"
                "\n\nReference:"
                f"\n\n{test_config.reference}"
                f"\n\nSchema description:\n  {json.dumps(schema_section, indent=4)}"
                "\n\nHint: Return a valid type, or document the value as nullable"
            )
        schema_section = normalize_schema_section(schema_section)
        if "oneOf" in schema_section:
            self.handle_one_of(
                schema_section=schema_section,
                data=data,
                reference=test_config.reference,
                test_config=test_config,
            )
            return
        if "anyOf" in schema_section:
            self.handle_any_of(
                schema_section=schema_section,
                data=data,
                reference=test_config.reference,
                test_config=test_config,
            )
            return

        schema_section_type = self.get_schema_type(schema_section)
        if not schema_section_type:
            return
        combined_validators = cast(
            "list[Callable[[dict, Any], Optional[str]]]",
            [
                validate_type,
                validate_format,
                validate_pattern,
                validate_multiple_of,
                validate_minimum,
                validate_maximum,
                validate_unique_items,
                validate_min_length,
                validate_max_length,
                validate_min_items,
                validate_max_items,
                validate_max_properties,
                validate_min_properties,
                validate_enum,
                *self.validators,
                *(test_config.validators or []),
            ],
        )
        for validator in combined_validators:
            error = validator(schema_section, data)
            if error:
                raise DocumentationError(
                    f"\n\n{error}"
                    "\n\nReference: "
                    f"\n\n{test_config.reference}"
                    f"\n\n {test_config.http_message.capitalize()} value:\n  {data}"
                    f"\n Schema description:\n  {schema_section}"
                )

        if schema_section_type == "object":
            self.test_openapi_object(
                schema_section=schema_section, data=data, test_config=test_config
            )
        elif schema_section_type == "array":
            self.test_openapi_array(
                schema_section=schema_section, data=data, test_config=test_config
            )

    def test_openapi_object(
        self,
        schema_section: dict,
        data: dict,
        test_config: OpenAPITestConfig,
    ) -> None:
        """
        1. Validate that casing is correct for both request/response and schema
        2. Check if any required key is missing from the request/response
        3. Check if any request/response key is not in the schema
        4. Validate sub-schema/nested data
        """
        properties = schema_section.get("properties", {})
        write_only_properties = [
            key for key in properties.keys() if properties[key].get("writeOnly")
        ]
        required_keys = [
            key
            for key in schema_section.get("required", [])
            if key not in write_only_properties
        ]
        request_response_keys = data.keys()
        additional_properties: bool | dict | None = schema_section.get(
            "additionalProperties"
        )
        additional_properties_allowed = additional_properties is not None
        if additional_properties_allowed and not isinstance(
            additional_properties, (bool, dict)
        ):
            raise OpenAPISchemaError("Invalid additionalProperties type")
        for key in properties.keys():
            self.test_key_casing(key, test_config.case_tester, test_config.ignore_case)
            if key in required_keys and key not in request_response_keys:
                raise DocumentationError(
                    f"{VALIDATE_MISSING_KEY_ERROR.format(missing_key=key, http_message=test_config.http_message)}"
                    "\n\nReference:"
                    f"\n\n{test_config.reference} > {key}"
                    f"\n\n{test_config.http_message.capitalize()} body:\n  {serialize_schema_section_data(data=data)}"
                    f"\nSchema section:\n  {serialize_schema_section_data(data=properties)}"
                    "\n\nHint: Remove the key from your OpenAPI docs, or"
                    f" include it in your API {test_config.http_message}"
                )
        for key in request_response_keys:
            self.test_key_casing(key, test_config.case_tester, test_config.ignore_case)
            if key not in properties and not additional_properties_allowed:
                raise DocumentationError(
                    f"{VALIDATE_EXCESS_KEY_ERROR.format(excess_key=key, http_message=test_config.http_message)}"
                    "\n\nReference:"
                    f"\n\n{test_config.reference} > {key}"
                    f"\n\n{test_config.http_message.capitalize()} body:\n  {serialize_schema_section_data(data=data)}"
                    f"\n\nSchema section:\n  {serialize_schema_section_data(data=properties)}"
                    "\n\nHint: Remove the key from your API"
                    f" {test_config.http_message}, or include it in your OpenAPI docs"
                )
            if key in write_only_properties:
                raise DocumentationError(
                    f"{VALIDATE_WRITE_ONLY_RESPONSE_KEY_ERROR.format(write_only_key=key)}\n\nReference:"
                    f"\n\n{test_config.reference} > {key}"
                    f"\n\n{test_config.http_message.capitalize()} body:\n  {serialize_schema_section_data(data=data)}"
                    f"\nSchema section:\n  {serialize_schema_section_data(data=properties)}"
                    f"\n\nHint: Remove the key from your API {test_config.http_message}, or"
                    ' remove the "WriteOnly" restriction'
                )
        for key, value in data.items():
            if key in properties:
                drill_down_test_config = copy(test_config)
                drill_down_test_config.reference = f"{test_config.reference} > {key}"
                self.test_schema_section(
                    schema_section=properties[key],
                    data=value,
                    test_config=drill_down_test_config,
                )
            elif isinstance(additional_properties, dict):
                drill_down_test_config = copy(test_config)
                drill_down_test_config.reference = f"{test_config.reference} > {key}"
                self.test_schema_section(
                    schema_section=additional_properties,
                    data=value,
                    test_config=drill_down_test_config,
                )

    def test_openapi_array(
        self, schema_section: dict[str, Any], data: dict, test_config: OpenAPITestConfig
    ) -> None:
        for array_item in data:
            array_item_test_config = copy(test_config)
            array_item_test_config.reference = f"{test_config.reference}"
            self.test_schema_section(
                # the items keyword is required in arrays
                schema_section=schema_section["items"],
                data=array_item,
                test_config=array_item_test_config,
            )

    def validate_request(
        self,
        response_handler: ResponseHandler,
        test_config: OpenAPITestConfig | None = None,
    ) -> None:
        """
        Verifies that an OpenAPI schema definition matches an API request body.

        :param request: The HTTP request
        :param case_tester: Optional Callable that checks a string's casing
        :param ignore_case: Optional list of keys to ignore in case testing
        :param validators: Optional list of validator functions
        :param **kwargs: Request keyword arguments
        :raises: ``openapi_tester.exceptions.DocumentationError`` for inconsistencies in the API response and schema.
                 ``openapi_tester.exceptions.CaseError`` for case errors.
        """
        if self.is_openapi_schema():
            # TODO: Implement for other schema types
            if test_config:
                test_config.http_message = "request"
            else:
                test_config = OpenAPITestConfig(
                    http_message="request",
                    reference=f"{response_handler.request.method} {response_handler.request.path} > request",
                )

            request_body_schema = self.get_request_body_schema_section(
                response_handler.request, test_config=test_config
            )

            if request_body_schema:
                self.test_schema_section(
                    schema_section=request_body_schema,
                    data=response_handler.request.data,
                    test_config=test_config,
                )

    def validate_response(
        self,
        response_handler: ResponseHandler,
        test_config: OpenAPITestConfig | None = None,
    ) -> None:
        """
        Verifies that an OpenAPI schema definition matches an API response.

        :param response: The HTTP response
        :param case_tester: Optional Callable that checks a string's casing
        :param ignore_case: Optional list of keys to ignore in case testing
        :param validators: Optional list of validator functions
        :raises: ``openapi_tester.exceptions.DocumentationError`` for inconsistencies in the API response and schema.
                 ``openapi_tester.exceptions.CaseError`` for case errors.
        """
        if test_config:
            test_config.http_message = "response"
        else:
            request = response_handler.request
            test_config = OpenAPITestConfig(
                http_message="response",
                reference=f"{request.method} {request.path} > response > {response_handler.response.status_code}",
            )
        response_schema = self.get_response_schema_section(
            response_handler, test_config=test_config
        )
        self.test_schema_section(
            schema_section=response_schema,
            data=response_handler.data,
            test_config=test_config,
        )



"""Schema Validators"""

from __future__ import annotations

import base64
import json
import re
from typing import TYPE_CHECKING
from uuid import UUID

from django.core.exceptions import ValidationError
from django.core.validators import (
    EmailValidator,
    URLValidator,
    validate_ipv4_address,
    validate_ipv6_address,
)
from django.utils.dateparse import parse_date, parse_datetime, parse_time

from openapi_tester.constants import (
    INTERNET_PROTOCOLS,
    INVALID_PATTERN_ERROR,
    NUMERIC_FORMATS,
    VALIDATE_ENUM_ERROR,
    VALIDATE_FORMAT_ERROR,
    VALIDATE_MAX_ARRAY_LENGTH_ERROR,
    VALIDATE_MAX_LENGTH_ERROR,
    VALIDATE_MAXIMUM_ERROR,
    VALIDATE_MAXIMUM_NUMBER_OF_PROPERTIES_ERROR,
    VALIDATE_MIN_ARRAY_LENGTH_ERROR,
    VALIDATE_MIN_LENGTH_ERROR,
    VALIDATE_MINIMUM_ERROR,
    VALIDATE_MINIMUM_NUMBER_OF_PROPERTIES_ERROR,
    VALIDATE_MULTIPLE_OF_ERROR,
    VALIDATE_PATTERN_ERROR,
    VALIDATE_TYPE_ERROR,
    VALIDATE_UNIQUE_ITEMS_ERROR,
)
from openapi_tester.exceptions import OpenAPISchemaError

if TYPE_CHECKING:
    from typing import Any, Callable


def create_validator(
    validation_fn: Callable, wrap_as_validator: bool = False
) -> Callable[[Any], bool]:
    def wrapped(value: Any) -> bool:
        try:
            return bool(validation_fn(value)) or not wrap_as_validator
        except (ValueError, ValidationError):
            return False

    return wrapped


number_format_validator = create_validator(
    lambda x: isinstance(x, float) if x != 0 else isinstance(x, (int, float)), True
)

base64_format_validator = create_validator(
    lambda x: base64.b64encode(base64.b64decode(x, validate=True)) == x
)

VALIDATOR_MAP: dict[str, Callable] = {
    # by type
    "string": create_validator(lambda x: isinstance(x, str), True),
    "file": create_validator(lambda x: isinstance(x, str), True),
    "boolean": create_validator(lambda x: isinstance(x, bool), True),
    "integer": create_validator(
        lambda x: isinstance(x, int) and not isinstance(x, bool), True
    ),
    "number": create_validator(
        lambda x: isinstance(x, (float, int)) and not isinstance(x, bool), True
    ),
    "object": create_validator(lambda x: isinstance(x, dict), True),
    "array": create_validator(lambda x: isinstance(x, list), True),
    # by format
    "byte": base64_format_validator,
    "base64": base64_format_validator,
    "date": create_validator(parse_date, True),
    "date-time": create_validator(parse_datetime, True),
    "double": number_format_validator,
    "email": create_validator(EmailValidator()),
    "float": number_format_validator,
    "ipv4": create_validator(validate_ipv4_address),
    "ipv6": create_validator(validate_ipv6_address),
    "time": create_validator(parse_time, True),
    "uri": create_validator(URLValidator()),
    "url": create_validator(URLValidator()),
    "uuid": create_validator(UUID),
}


def validate_type(schema_section: dict[str, Any], data: Any) -> str | None:
    an_articles = ["integer", "object", "array"]
    schema_types: str = schema_section.get("type", "object")
    if isinstance(schema_types, list):
        has_type = False
        for schema_type in schema_types:
            if VALIDATOR_MAP[schema_type](data):
                return None
        if not has_type:
            return VALIDATE_TYPE_ERROR.format(
                article="a" if schema_types not in an_articles else "an",
                type=schema_types,
                received=f'"{data}"' if isinstance(data, str) else data,
            )
    if not VALIDATOR_MAP[schema_types](data):
        return VALIDATE_TYPE_ERROR.format(
            article="a" if schema_types not in an_articles else "an",
            type=schema_types,
            received=f'"{data}"' if isinstance(data, str) else data,
        )
    return None


def validate_format(schema_section: dict[str, Any], data: Any) -> str | None:
    value = data
    schema_format: str = schema_section.get("format", "")
    schema_type: str = schema_section.get("type", "object")
    if schema_format in VALIDATOR_MAP:
        if not isinstance(schema_type, list) and schema_type == "string":
            try:
                if schema_format == "integer":
                    value = int(data)
                if schema_format in NUMERIC_FORMATS:
                    value = float(data)
            except ValueError:
                return VALIDATE_FORMAT_ERROR.format(
                    article="an" if format in INTERNET_PROTOCOLS else "a",
                    format=schema_format,
                    type=schema_type,
                    received=f'"{value}"',
                )
        if not VALIDATOR_MAP[schema_format](value):
            return VALIDATE_FORMAT_ERROR.format(
                article="an" if format in INTERNET_PROTOCOLS else "a",
                format=schema_format,
                type=schema_type,
                received=f'"{value}"',
            )
    return None


def validate_enum(schema_section: dict[str, Any], data: Any) -> str | None:
    enum = schema_section.get("enum")
    if enum and data not in enum:
        return VALIDATE_ENUM_ERROR.format(
            enum=schema_section["enum"], received=f'"{data}"'
        )
    return None


def validate_pattern(schema_section: dict[str, Any], data: str) -> str | None:
    pattern = schema_section.get("pattern")
    if not pattern:
        return None
    try:
        compiled_pattern = re.compile(pattern)
    except re.error as e:
        raise OpenAPISchemaError(INVALID_PATTERN_ERROR.format(pattern=pattern)) from e
    if not compiled_pattern.match(str(data)):
        return VALIDATE_PATTERN_ERROR.format(data=data, pattern=pattern)
    return None


def validate_multiple_of(
    schema_section: dict[str, Any], data: int | float
) -> str | None:
    multiple = schema_section.get("multipleOf")
    if multiple and data % multiple != 0:
        return VALIDATE_MULTIPLE_OF_ERROR.format(data=data, multiple=multiple)
    return None


def validate_maximum(schema_section: dict[str, Any], data: int | float) -> str | None:
    maximum = schema_section.get("maximum")
    exclusive_maximum = schema_section.get("exclusiveMaximum")
    if maximum and exclusive_maximum and data >= maximum:
        return VALIDATE_MAXIMUM_ERROR.format(data=data, maximum=maximum - 1)
    if maximum and not exclusive_maximum and data > maximum:
        return VALIDATE_MAXIMUM_ERROR.format(data=data, maximum=maximum)
    return None


def validate_minimum(schema_section: dict[str, Any], data: int | float) -> str | None:
    minimum = schema_section.get("minimum")
    exclusive_minimum = schema_section.get("exclusiveMinimum")
    if minimum and exclusive_minimum and data <= minimum:
        return VALIDATE_MINIMUM_ERROR.format(data=data, minimum=minimum + 1)
    if minimum and not exclusive_minimum and data < minimum:
        return VALIDATE_MINIMUM_ERROR.format(data=data, minimum=minimum)
    return None


def validate_unique_items(
    schema_section: dict[str, Any], data: list[Any]
) -> str | None:
    unique_items = schema_section.get("uniqueItems")
    if unique_items:
        comparison_data = (
            json.dumps(item, sort_keys=True) if isinstance(item, dict) else item
            for item in data
        )
        if len(set(comparison_data)) != len(data):
            return VALIDATE_UNIQUE_ITEMS_ERROR.format(data=data)
    return None


def validate_min_length(schema_section: dict[str, Any], data: str) -> str | None:
    min_length: int | None = schema_section.get("minLength")
    if min_length and len(data) < min_length:
        return VALIDATE_MIN_LENGTH_ERROR.format(data=data, min_length=min_length)
    return None


def validate_max_length(schema_section: dict[str, Any], data: str) -> str | None:
    max_length: int | None = schema_section.get("maxLength")
    if max_length and len(data) > max_length:
        return VALIDATE_MAX_LENGTH_ERROR.format(data=data, max_length=max_length)
    return None


def validate_min_items(schema_section: dict[str, Any], data: list) -> str | None:
    min_length: int | None = schema_section.get("minItems")
    if min_length and len(data) < min_length:
        return VALIDATE_MIN_ARRAY_LENGTH_ERROR.format(data=data, min_length=min_length)
    return None


def validate_max_items(schema_section: dict[str, Any], data: list) -> str | None:
    max_length: int | None = schema_section.get("maxItems")
    if max_length and len(data) > max_length:
        return VALIDATE_MAX_ARRAY_LENGTH_ERROR.format(data=data, max_length=max_length)
    return None


def validate_min_properties(schema_section: dict[str, Any], data: dict) -> str | None:
    min_properties: int | None = schema_section.get("minProperties")
    if min_properties and len(data.keys()) < int(min_properties):
        return VALIDATE_MINIMUM_NUMBER_OF_PROPERTIES_ERROR.format(
            data=data, min_length=min_properties
        )
    return None


def validate_max_properties(schema_section: dict[str, Any], data: dict) -> str | None:
    max_properties: int | None = schema_section.get("maxProperties")
    if max_properties and len(data.keys()) > int(max_properties):
        return VALIDATE_MAXIMUM_NUMBER_OF_PROPERTIES_ERROR.format(
            data=data, max_length=max_properties
        )
    return None



"""Subclass of ``APIClient`` using ``SchemaTester`` to validate responses."""

from __future__ import annotations

import http
from typing import TYPE_CHECKING

# pylint: disable=import-error
from ninja import NinjaAPI, Router
from ninja.testing import TestClient
from rest_framework.test import APIClient

from .response_handler_factory import ResponseHandlerFactory
from .schema_tester import SchemaTester
from .utils import serialize_json

if TYPE_CHECKING:
    from rest_framework.response import Response


class OpenAPIClient(APIClient):
    """``APIClient`` validating responses against OpenAPI schema."""

    def __init__(
        self,
        *args,
        schema_tester: SchemaTester | None = None,
        **kwargs,
    ) -> None:
        """Initialize ``OpenAPIClient`` instance."""
        super().__init__(*args, **kwargs)
        self.schema_tester = schema_tester or self._schema_tester_factory()

    def request(self, *args, **kwargs) -> Response:  # type: ignore[override]
        """Validate fetched response against given OpenAPI schema."""
        response = super().request(**kwargs)
        response_handler = ResponseHandlerFactory.create(
            *args, response=response, **kwargs
        )
        if self._is_successful_response(response):
            self.schema_tester.validate_request(response_handler=response_handler)
        self.schema_tester.validate_response(response_handler=response_handler)
        return response

    # pylint: disable=W0622
    @serialize_json
    def post(
        self,
        *args,
        content_type="application/json",
        **kwargs,
    ):
        return super().post(
            *args,
            content_type=content_type,
            **kwargs,
        )

    # pylint: disable=W0622
    @serialize_json
    def put(
        self,
        *args,
        content_type="application/json",
        **kwargs,
    ):
        return super().put(
            *args,
            content_type=content_type,
            **kwargs,
        )

    # pylint: disable=W0622
    @serialize_json
    def patch(self, *args, content_type="application/json", **kwargs):
        return super().patch(
            *args,
            content_type=content_type,
            **kwargs,
        )

    # pylint: disable=W0622
    @serialize_json
    def delete(
        self,
        *args,
        content_type="application/json",
        **kwargs,
    ):
        return super().delete(
            *args,
            content_type=content_type,
            **kwargs,
        )

    # pylint: disable=W0622
    @serialize_json
    def options(
        self,
        *args,
        content_type="application/json",
        **kwargs,
    ):
        return super().options(
            *args,
            content_type=content_type,
            **kwargs,
        )

    @staticmethod
    def _is_successful_response(response: Response) -> bool:
        return response.status_code < http.HTTPStatus.BAD_REQUEST

    @staticmethod
    def _schema_tester_factory() -> SchemaTester:
        """Factory of default ``SchemaTester`` instances."""
        return SchemaTester()


# pylint: disable=R0903
class OpenAPINinjaClient(TestClient):
    """``APINinjaClient`` validating responses against OpenAPI schema."""

    def __init__(
        self,
        *args,
        router_or_app: NinjaAPI | Router,
        path_prefix: str = "",
        schema_tester: SchemaTester | None = None,
        **kwargs,
    ) -> None:
        """Initialize ``OpenAPIClient`` instance."""
        super().__init__(*args, router_or_app=router_or_app, **kwargs)
        self.schema_tester = schema_tester or self._schema_tester_factory()
        self._ninja_path_prefix = path_prefix

    def request(self, *args, **kwargs) -> Response:
        """Validate fetched response against given OpenAPI schema."""
        response = super().request(*args, **kwargs)
        response_handler = ResponseHandlerFactory.create(
            *args, response=response, path_prefix=self._ninja_path_prefix, **kwargs
        )
        if self._is_successful_response(response):
            self.schema_tester.validate_request(response_handler=response_handler)
        self.schema_tester.validate_response(response_handler)
        return response

    @staticmethod
    def _is_successful_response(response: Response) -> bool:
        return response.status_code < http.HTTPStatus.BAD_REQUEST

    @staticmethod
    def _schema_tester_factory() -> SchemaTester:
        """Factory of default ``SchemaTester`` instances."""
        return SchemaTester()



"""Loaders Module"""

from __future__ import annotations

import difflib
import json
import pathlib
import re
from json import dumps, loads
from typing import TYPE_CHECKING, cast
from urllib.parse import urlparse

import requests
import yaml
from django.urls import Resolver404, resolve
from django.utils.functional import cached_property
from openapi_spec_validator import (
    OpenAPIV2SpecValidator,
    OpenAPIV30SpecValidator,
    OpenAPIV31SpecValidator,
)
from prance.util.resolver import RefResolver
from rest_framework.schemas.generators import BaseSchemaGenerator, EndpointEnumerator
from rest_framework.settings import api_settings

from openapi_tester.constants import UNDOCUMENTED_SCHEMA_SECTION_ERROR
from openapi_tester.exceptions import UndocumentedSchemaSectionError

if TYPE_CHECKING:
    from typing import Any, Callable
    from urllib.parse import ParseResult

    from django.urls import ResolverMatch
    from rest_framework.views import APIView


def handle_recursion_limit(schema: dict) -> Callable:
    """
    We are using a currying pattern to pass schema into the scope of the handler.
    """

    # noinspection PyUnusedLocal
    def handler(iteration: int, parse_result: ParseResult, recursions: tuple):  # pylint: disable=unused-argument
        fragment = parse_result.fragment
        keys = [key for key in fragment.split("/") if key]
        definition = schema
        for key in keys:
            definition = definition[key]
        return definition

    return handler


class BaseSchemaLoader:
    """
    Base class for OpenAPI schema loading classes.

    Contains a template of methods that are required from a loader class, and a range of helper methods for interacting
    with an OpenAPI schema.
    """

    base_path = "/"
    field_key_map: dict[str, str]
    schema: dict | None = None

    def __init__(self, field_key_map: dict[str, str] | None = None):
        super().__init__()
        self.schema: dict | None = None
        self.field_key_map = field_key_map or {}

    def load_schema(self) -> dict:
        """
        Put logic required to load a schema and return it here.
        """
        raise NotImplementedError("The `load_schema` method has to be overwritten.")

    def get_schema(self) -> dict:
        """
        Returns OpenAPI schema.
        """
        if self.schema:
            return self.schema
        self.set_schema(self.load_schema())
        return self.get_schema()

    def de_reference_schema(self, schema: dict) -> dict:
        url = schema.get("basePath", self.base_path)
        recursion_handler = handle_recursion_limit(schema)
        resolver = RefResolver(
            schema,
            recursion_limit_handler=recursion_handler,
            recursion_limit=10,
            url=url,
        )
        resolver.resolve_references()
        return resolver.specs

    def normalize_schema_paths(self, schema: dict) -> dict[str, dict]:
        normalized_paths: dict[str, dict] = {}
        for key, value in schema["paths"].items():
            try:
                parameterized_path, _ = self.resolve_path(
                    endpoint_path=key, method=list(value.keys())[0]
                )
                normalized_paths[parameterized_path] = value
            except ValueError:
                normalized_paths[key] = value
        return {**schema, "paths": normalized_paths}

    @staticmethod
    def validate_schema(schema: dict):
        if "openapi" in schema:
            openapi_version_pattern = re.compile(r"^(\d)\.(\d+)")
            result = openapi_version_pattern.findall(schema["openapi"])
            if result:
                major, minor = result[0]
                if (major, minor) == ("3", "0"):
                    validator = OpenAPIV30SpecValidator(schema=schema)
                elif (major, minor) == ("3", "1"):
                    validator = OpenAPIV31SpecValidator(schema=schema)
                else:
                    raise UndocumentedSchemaSectionError(
                        UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(
                            key=schema["openapi"],
                            error_addon="Support might need to be added.",
                        )
                    )
            else:
                raise UndocumentedSchemaSectionError(
                    UNDOCUMENTED_SCHEMA_SECTION_ERROR.format(key=schema["openapi"])
                )
        else:
            validator = OpenAPIV2SpecValidator(schema=schema)
        validator.validate()

    def set_schema(self, schema: dict) -> None:
        """
        Sets self.schema and self.original_schema.
        """
        de_referenced_schema = self.de_reference_schema(schema)
        self.validate_schema(de_referenced_schema)

        self.schema = self.normalize_schema_paths(de_referenced_schema)

    @cached_property
    def endpoints(self) -> list[str]:
        """
        Returns a list of endpoint paths.
        """
        return list(
            {endpoint[0] for endpoint in EndpointEnumerator().get_api_endpoints()}
        )

    def resolve_path(
        self, endpoint_path: str, method: str
    ) -> tuple[str, ResolverMatch]:
        """
        Resolves a Django path.
        """
        url_object = urlparse(endpoint_path)
        parsed_path = url_object.path
        if not parsed_path.startswith("/"):
            parsed_path = "/" + parsed_path
        for key, value in self.field_key_map.items():
            if value != "pk" and key in parsed_path:
                parsed_path = parsed_path.replace(f"{{{key}}}", value)
        for path in [parsed_path, parsed_path[:-1]]:
            try:
                resolved_route = resolve(path)
            except Resolver404:
                continue
            else:
                for key, value in reversed(list(resolved_route.kwargs.items())):
                    index = path.rfind(str(value))
                    path = f"{path[:index]}{{{key}}}{path[index + len(str(value)):]}"
                if "{pk}" in path and api_settings.SCHEMA_COERCE_PATH_PK:  # noqa: FS003
                    path, resolved_route = self.handle_pk_parameter(
                        resolved_route=resolved_route, path=path, method=method
                    )
                return path, resolved_route
        message = f"Could not resolve path `{endpoint_path}`."
        close_matches = difflib.get_close_matches(endpoint_path, self.endpoints)
        if close_matches:
            message += "\n\nDid you mean one of these?\n\n- " + "\n- ".join(
                close_matches
            )
        raise ValueError(message)

    @staticmethod
    def handle_pk_parameter(
        resolved_route: ResolverMatch, path: str, method: str
    ) -> tuple[str, ResolverMatch]:
        """
        Handle the DRF conversion of params called {pk} into a named parameter based on Model field
        """
        coerced_path = BaseSchemaGenerator().coerce_path(
            path=path, method=method, view=cast("APIView", resolved_route.func)
        )
        pk_field_name = "".join(
            entry.replace("+ ", "")
            for entry in difflib.Differ().compare(path, coerced_path)
            if "+ " in entry
        )
        resolved_route.kwargs[pk_field_name] = resolved_route.kwargs["pk"]
        del resolved_route.kwargs["pk"]
        return coerced_path, resolved_route


class DrfYasgSchemaLoader(BaseSchemaLoader):
    """
    Loads OpenAPI schema generated by drf_yasg.
    """

    def __init__(self, field_key_map: dict[str, str] | None = None) -> None:
        super().__init__(field_key_map=field_key_map)
        from drf_yasg.generators import OpenAPISchemaGenerator
        from drf_yasg.openapi import Info

        self.schema_generator = OpenAPISchemaGenerator(
            info=Info(title="", default_version="")
        )

    def load_schema(self) -> dict:
        """
        Loads generated schema from drf-yasg and returns it as a dict.
        """
        odict_schema = self.schema_generator.get_schema(None, True)
        return cast("dict", loads(dumps(odict_schema.as_odict())))

    def resolve_path(
        self, endpoint_path: str, method: str
    ) -> tuple[str, ResolverMatch]:
        de_parameterized_path, resolved_path = super().resolve_path(
            endpoint_path=endpoint_path, method=method
        )
        path_prefix = self.schema_generator.determine_path_prefix(self.endpoints)
        trim_length = len(path_prefix) if path_prefix != "/" else 0
        return de_parameterized_path[trim_length:], resolved_path


class DrfSpectacularSchemaLoader(BaseSchemaLoader):
    """
    Loads OpenAPI schema generated by drf_spectacular.
    """

    def __init__(self, field_key_map: dict[str, str] | None = None) -> None:
        super().__init__(field_key_map=field_key_map)
        from drf_spectacular.generators import SchemaGenerator

        self.schema_generator = SchemaGenerator()

    def load_schema(self) -> dict:
        """
        Loads generated schema from drf_spectacular and returns it as a dict.
        """
        return cast("dict", loads(dumps(self.schema_generator.get_schema(public=True))))

    def resolve_path(
        self, endpoint_path: str, method: str
    ) -> tuple[str, ResolverMatch]:
        from drf_spectacular.settings import spectacular_settings

        de_parameterized_path, resolved_path = super().resolve_path(
            endpoint_path=endpoint_path, method=method
        )
        return (
            de_parameterized_path[len(spectacular_settings.SCHEMA_PATH_PREFIX or "") :],
            resolved_path,
        )


class StaticSchemaLoader(BaseSchemaLoader):
    """
    Loads OpenAPI schema from a static file.
    """

    def __init__(self, path: str, field_key_map: dict[str, str] | None = None):
        super().__init__(field_key_map=field_key_map)

        self.path = path if not isinstance(path, pathlib.PosixPath) else str(path)

    def load_schema(self) -> dict[str, Any]:
        """
        Loads a static OpenAPI schema from file, and parses it to a python dict.

        :return: Schema contents as a dict
        :raises: ImproperlyConfigured
        """
        with open(self.path, encoding="utf-8") as file:
            content = file.read()
            return cast(
                "dict",
                json.loads(content)
                if ".json" in self.path
                else yaml.load(content, Loader=yaml.FullLoader),
            )


class UrlStaticSchemaLoader(BaseSchemaLoader):
    """
    Loads OpenAPI schema from an url static file.
    """

    def __init__(self, url: str, field_key_map: dict[str, str] | None = None):
        super().__init__(field_key_map=field_key_map)
        self.url = url

    def load_schema(self) -> dict[str, Any]:
        """
        Loads a static OpenAPI schema from url, and parses it to a python dict.

        :return: Schema contents as a dict
        :raises: ImproperlyConfigured
        """
        response = requests.get(self.url, timeout=20)
        return cast(
            "dict",
            (
                json.loads(response.content)
                if ".json" in self.url
                else yaml.load(response.content, Loader=yaml.FullLoader)
            ),
        )



"""Case testers - this module includes helper functions to test key casing"""

from __future__ import annotations

from typing import TYPE_CHECKING

from inflection import camelize, dasherize, underscore

from openapi_tester.exceptions import CaseError

if TYPE_CHECKING:
    from typing import Any, Callable


def _create_tester(casing: str, handler: Callable[[Any], str]) -> Callable[[str], None]:
    """factory function for creating testers"""

    def tester(key: str) -> None:
        stripped = key.strip()
        if stripped and handler(stripped) != stripped:
            raise CaseError(key=key, case=casing, expected=handler(key))

    return tester


def _camelize(string: str) -> str:
    return camelize(underscore(string), False)


def _pascalize(string: str) -> str:
    return camelize(underscore(string))


def _kebabize(string: str) -> str:
    return dasherize(underscore(string))


is_camel_case = _create_tester("camelCased", _camelize)
is_kebab_case = _create_tester("kebab-cased", _kebabize)
is_pascal_case = _create_tester("PascalCased", _pascalize)
is_snake_case = _create_tester("snake_cased", underscore)
